@@ -13,10 +13,21 @@
 #endif
 
 Display		*dpy;
-Client		*current = NULL;
+/*
 int		screen;
 Window		root;
 GC		invert_gc;
+#ifdef VWM
+XColor		fg, bg, fc;
+#else
+XColor		fg, bg;
+#endif
+*/
+int 		num_screens;
+ScreenInfo	*screens;
+ScreenInfo	*current_screen;
+Client		*current = NULL;
+Window		initialising = None;
 XFontStruct	*font;
 Client		*head_client;
 Atom		xa_wm_state;
@@ -30,28 +41,20 @@ char		*opt_display = "";
 char		*opt_font = DEF_FONT;
 char		*opt_fg = DEF_FG;
 char		*opt_bg = DEF_BG;
-char		*opt_term;
+char		**opt_term = NULL;
 int		opt_bw = DEF_BW;
 #ifdef VWM
-XColor		fg, bg, fc;
 char		*opt_fc = DEF_FC;
-# ifndef VDESK_BOTH
- int		vdesk = XK_1;
-# endif
-#else
-XColor		fg, bg;
-#endif /* def VWM */
-#ifdef SANITY
-int		tracked_count = 0;
+int		vdesk = KEY_TO_VDESK(XK_1);
 #endif
 #ifdef SHAPE
 int		have_shape, shape_event;
 #endif
-#ifdef VDESK_BOTH
-Atom	vdesk_desktop;
-char * vdesk_property_name = DEF_VDESK_PROP;
-#endif
 int		quitting = 0;
+#ifdef MWM_HINTS
+Atom		mwm_hints;
+#endif
+unsigned int numlockmask = 0;
 
 int main(int argc, char *argv[]) {
 	struct sigaction act;
@@ -61,8 +64,9 @@ int main(int argc, char *argv[]) {
 	for (i = 1; i < argc; i++) {
 		if (!strcmp(argv[i], "-fn") && i+1<argc)
 			opt_font = argv[++i];
-		else if (!strcmp(argv[i], "-display") && i+1<argc)
+		else if (!strcmp(argv[i], "-display") && i+1<argc) {
 			opt_display = argv[++i];
+		}
 		else if (!strcmp(argv[i], "-fg") && i+1<argc)
 			opt_fg = argv[++i];
 		else if (!strcmp(argv[i], "-bg") && i+1<argc)
@@ -70,15 +74,14 @@ int main(int argc, char *argv[]) {
 #ifdef VWM
 		else if (!strcmp(argv[i], "-fc") && i+1<argc)
 			opt_fc = argv[++i];
-#endif
-#ifdef VDESK_BOTH
-		else if (!strcmp(argv[i], "-deskprop") && i+1<argc)
-			vdesk_property_name = argv[++i];
 #endif
 		else if (!strcmp(argv[i], "-bw") && i+1<argc)
 			opt_bw = atoi(argv[++i]);
 		else if (!strcmp(argv[i], "-term") && i+1<argc) {
-			opt_term = argv[++i];
+			opt_term = (char **)malloc(3 * sizeof(char *));
+			opt_term[0] = argv[++i];
+			opt_term[1] = opt_term[0];
+			opt_term[2] = NULL;
 #ifdef STDIO
 		} else if (!strcmp(argv[i], "-V")) {
 			printf("evilwm version " VERSION "\n");
@@ -88,13 +91,16 @@ int main(int argc, char *argv[]) {
 #ifdef STDIO
 			printf("usage: evilwm [-display display] [-term termprog] [-fg foreground]\n");
 			printf("\t[-bg background] [-bw borderwidth] [-V]\n");
-# ifdef VDESK_BOTH
-			printf("\t[-deskprop name]\n");
-# endif
 #endif
 			exit(2);
 		}
 	}
+	if (!opt_term) {
+		opt_term = (char **)malloc(3 * sizeof(char *));
+		opt_term[0] = DEF_TERM;
+		opt_term[1] = opt_term[0];
+		opt_term[2] = NULL;
+	}
 
 	act.sa_handler = handle_signal;
 	sigemptyset(&act.sa_mask);
@@ -114,13 +120,8 @@ int main(int argc, char *argv[]) {
 	have_shape = XShapeQueryExtension(dpy, &shape_event, &i);
 #endif
 
-	scan_windows();
-
 	/* main event loop here */
 	for (;;) {
-#ifdef SANITY
-		sanity_check();
-#endif
 		XNextEvent(dpy, &ev);
 		switch (ev.type) {
 			case KeyPress:
@@ -143,6 +144,8 @@ int main(int argc, char *argv[]) {
 #endif
 			case EnterNotify:
 				handle_enter_event(&ev.xcrossing); break;
+			case LeaveNotify:
+				handle_leave_event(&ev.xcrossing); break;
 			case PropertyNotify:
 				handle_property_change(&ev.xproperty); break;
 			case UnmapNotify:
@@ -164,8 +167,9 @@ void setup_display() {
 	XGCValues gv;
 	XSetWindowAttributes attr;
 	XColor dummy;
-	int *i;
-	int keys_to_grab[] = {
+	XModifierKeymap *modmap;
+	KeySym *keysym;
+	KeySym keys_to_grab[] = {
 		KEY_NEW, KEY_KILL,
 		KEY_TOPLEFT, KEY_TOPRIGHT, KEY_BOTTOMLEFT, KEY_BOTTOMRIGHT,
 		KEY_LEFT, KEY_RIGHT, KEY_DOWN, KEY_UP,
@@ -176,10 +180,11 @@ void setup_display() {
 #endif
 		0
 	};
+	/* used in scanning windows (XQueryTree) */
+	unsigned int i, j, nwins;
+	Window dw1, dw2, *wins;
+	XWindowAttributes winattr;
 
-#ifdef XDEBUG
-	fprintf(stderr, "main:XOpenDisplay(); ");
-#endif
 	dpy = XOpenDisplay(opt_display);
 	if (!dpy) { 
 #ifdef STDIO
@@ -189,9 +194,6 @@ void setup_display() {
 	}
 	XSetErrorHandler(handle_xerror);
 
-	screen = DefaultScreen(dpy);
-	root = RootWindow(dpy, screen);
-
 	xa_wm_state = XInternAtom(dpy, "WM_STATE", False);
 	xa_wm_change_state = XInternAtom(dpy, "WM_CHANGE_STATE", False);
 	xa_wm_protos = XInternAtom(dpy, "WM_PROTOCOLS", False);
@@ -199,15 +201,8 @@ void setup_display() {
 #ifdef COLOURMAP
 	xa_wm_cmapwins = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
 #endif
-#ifdef VDESK_BOTH
-	vdesk_desktop = XInternAtom(dpy, vdesk_property_name, False);
-	SET_VDESK(1);
-#endif
-
-	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_fg, &fg, &dummy);
-	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_bg, &bg, &dummy);
-#ifdef VWM
-	XAllocNamedColor(dpy, DefaultColormap(dpy, screen), opt_fc, &fc, &dummy);
+#ifdef MWM_HINTS
+	mwm_hints = XInternAtom(dpy, _XA_MWM_HINTS, False);
 #endif
 
 	font = XLoadQueryFont(dpy, opt_font);
@@ -216,13 +211,29 @@ void setup_display() {
 	move_curs = XCreateFontCursor(dpy, XC_fleur);
 	resize_curs = XCreateFontCursor(dpy, XC_plus);
 
+	/* find out which modifier is NumLock - we'll use this when grabbing
+	 * every combination of modifiers we can think of */
+	modmap = XGetModifierMapping(dpy);
+	for (i = 0; i < 8; i++) {
+		for (j = 0; j < modmap->max_keypermod; j++) {
+			if (modmap->modifiermap[i*modmap->max_keypermod+j] == XKeysymToKeycode(dpy, XK_Num_Lock)) {
+				numlockmask = (1<<i);
+#ifdef DEBUG
+				fprintf(stderr, "setup_display() : XK_Num_Lock is (1<<0x%02x)\n", i);
+#endif
+			}
+		}
+	}
+	XFree(modmap);
+
+	/* set up GC parameters - same for each screen */
 	gv.function = GXinvert;
 	gv.subwindow_mode = IncludeInferiors;
 	gv.line_width = 1;  /* opt_bw */
 	gv.font = font->fid;
-	invert_gc = XCreateGC(dpy, root, GCFunction | GCSubwindowMode | GCLineWidth | GCFont, &gv);
 
-	attr.event_mask = ChildMask | PropertyChangeMask
+	/* set up root window attributes - same for each screen */
+	attr.event_mask = ChildMask | PropertyChangeMask | EnterWindowMask | LeaveWindowMask
 #ifdef COLOURMAP
 		| ColormapChangeMask
 #endif
@@ -230,34 +241,62 @@ void setup_display() {
 		| ButtonMask
 #endif
 		;
-	XChangeWindowAttributes(dpy, root, CWEventMask, &attr);
-	/* Unfortunately grabbing AnyKey under Solaris seems not to work */
-	/* XGrabKey(dpy, AnyKey, ControlMask|Mod1Mask, root, True, GrabModeAsync, GrabModeAsync); */
-	/* So now I grab each and every one. */
-	for (i = keys_to_grab; *i; i++) {
-		XGrabKey(dpy, XKeysymToKeycode(dpy, *i), ControlMask|Mod1Mask, root, True, GrabModeAsync, GrabModeAsync);
-		XGrabKey(dpy, XKeysymToKeycode(dpy, *i), Mod3Mask|ControlMask|Mod1Mask, root, True, GrabModeAsync, GrabModeAsync);
-	}
-	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_Tab), Mod1Mask, root, True, GrabModeAsync, GrabModeAsync);
-	XGrabKey(dpy, XKeysymToKeycode(dpy, XK_Tab), Mod3Mask|Mod1Mask, root, True, GrabModeAsync, GrabModeAsync);
-}
 
-void scan_windows() {
-	unsigned int i, nwins;
-	Window dw1, dw2, *wins;
-	XWindowAttributes attr;
+	/* now set up each screen in turn */
+	num_screens = ScreenCount(dpy);
+	screens = (ScreenInfo *)malloc(num_screens * sizeof(ScreenInfo));
+	for (i = 0; i < num_screens; i++) {
+		char *ds, *colon, *dot;
+		ds = DisplayString(dpy);
+		/* set up DISPLAY environment variable to use */
+		colon = rindex(ds, ':');
+		if (colon && num_screens > 1) {
+			screens[i].display = (char *)malloc(14 + strlen(ds));
+			strcpy(screens[i].display, "DISPLAY=");
+			strcat(screens[i].display, ds);
+			colon = rindex(screens[i].display, ':');
+			dot = index(colon, '.');
+			if (!dot)
+				dot = colon + strlen(colon);
+			snprintf(dot, 5, ".%d", i);
+		} else
+			screens[i].display = NULL;
 
-#ifdef XDEBUG
-	fprintf(stderr, "main:XQueryTree(); ");
+		screens[i].screen = i;
+		screens[i].root = RootWindow(dpy, i);
+
+		XAllocNamedColor(dpy, DefaultColormap(dpy, i), opt_fg, &screens[i].fg, &dummy);
+		XAllocNamedColor(dpy, DefaultColormap(dpy, i), opt_bg, &screens[i].bg, &dummy);
+#ifdef VWM
+		XAllocNamedColor(dpy, DefaultColormap(dpy, i), opt_fc, &screens[i].fc, &dummy);
 #endif
-	XQueryTree(dpy, root, &dw1, &dw2, &wins, &nwins);
+
+		screens[i].invert_gc = XCreateGC(dpy, screens[i].root, GCFunction | GCSubwindowMode | GCLineWidth | GCFont, &gv);
+
+		XChangeWindowAttributes(dpy, screens[i].root, CWEventMask, &attr);
+		/* Unfortunately grabbing AnyKey under Solaris seems not to work */
+		/* XGrabKey(dpy, AnyKey, ControlMask|Mod1Mask, root, True, GrabModeAsync, GrabModeAsync); */
+		/* So now I grab each and every one. */
+
+		for (keysym = keys_to_grab; *keysym; keysym++) {
+			grab_keysym(screens[i].root, ControlMask|Mod1Mask, *keysym);
+		}
+		grab_keysym(screens[i].root, Mod1Mask, XK_Tab);
+
+		/* scan all the windows on this screen */
 #ifdef XDEBUG
-	fprintf(stderr, "%d windows\n", nwins);
+		fprintf(stderr, "main:XQueryTree(); ");
 #endif
-	for (i = 0; i < nwins; i++) {
-		XGetWindowAttributes(dpy, wins[i], &attr);
-		if (!attr.override_redirect && attr.map_state == IsViewable)
-			make_new_client(wins[i]);
+		XQueryTree(dpy, screens[i].root, &dw1, &dw2, &wins, &nwins);
+#ifdef XDEBUG
+		fprintf(stderr, "%d windows\n", nwins);
+#endif
+		for (j = 0; j < nwins; j++) {
+			XGetWindowAttributes(dpy, wins[j], &winattr);
+			if (!winattr.override_redirect && winattr.map_state == IsViewable)
+				make_new_client(wins[j], &screens[i]);
+		}
+		XFree(wins);
 	}
-	XFree(wins);
+	current_screen = find_screen(DefaultScreen(dpy));
 }
